"""
Handle finding and parsing configuration file
"""
import configparser
import logging
from os import access, R_OK, X_OK
from pathlib import Path

log = logging.getLogger(__name__)

__mysql_tracer_ini = 'mysql-tracer.ini'
__user_config_path = Path.home().joinpath('.config', 'mysql-tracer', __mysql_tracer_ini)


def __find_config(config_name):
    """
    Search for a configuration file.

    First it searches current directory and above and then at user configuration folder

    :param config_name: name of the configuration file
    :return: path to the configuration file if it is found, else None
    :rtype Path | NoneType
    """
    current_dir = Path.cwd()
    while access(str(current_dir), X_OK) and current_dir.parent is not current_dir:
        config_path = current_dir.joinpath(config_name)
        log.debug('Searching configuration file at %s', config_path)
        if config_path.exists():
            log.debug('Found configuration file %s', config_path)
            if access(str(config_path), R_OK):
                return config_path
            else:
                raise PermissionError(config_path)
        current_dir = current_dir.parent

    log.debug('Searching configuration file at %s', __user_config_path)
    if __user_config_path.exists():
        if access(str(__user_config_path), R_OK):
            log.debug('Found configuration file %s', __user_config_path)
            return __user_config_path
        else:
            raise PermissionError(__user_config_path)

    log.debug('Did not find configuration file')
    return None


def get():
    config_path = __find_config(__mysql_tracer_ini)
    if config_path is None:
        return dict()

    config_parser = configparser.ConfigParser()
    config_parser.read(config_path)
    return dict(config_parser.items('mysql_tracer'))



"""
CLI script to run MySQL SELECT statements

It produces a copy of provided file with additional metadata and an export of results in CSV format
"""
import argparse


def get_main_args_parser(parents, defaults):
    """
    Parser for the arguments required to run mysql_tracer.

    :param parents: a list of parent argument parsers. They must NOT define a help option or else they will clash.
    :param defaults: a dictionary with default values. All actions with default values are set to not required.
    :return: an argument parser
    """
    parser = argparse.ArgumentParser(parents=parents + [get_database_args_parser()],
                                     description=__doc__,
                                     formatter_class=argparse.RawTextHelpFormatter)

    parser.set_defaults(**{
        'port': 3306,
        **defaults
    })

    query = parser.add_argument_group(title='Queries')
    query.add_argument('query', nargs='+', help='Path to a file containing a single sql statement')

    export = parser.add_argument_group(title='Export')
    excl_actions = export.add_mutually_exclusive_group()
    excl_actions.add_argument('-d', '--destination', help='Directory where to export results')
    excl_actions.add_argument('--display', action='store_true', help='Do not export results but display them to stdout')

    # noinspection PyProtectedMember
    for action in parser._actions:
        if action.required and action.default is not None:
            action.required = False

    return parser


def get_database_args_parser():
    parser = argparse.ArgumentParser(add_help=False)

    db = parser.add_argument_group(title='Database')
    db.add_argument('--host', required=True, help='MySQL server host')
    db.add_argument('--port', type=int, help='MySQL server port')
    db.add_argument('--user', required=True, help='MySQL database user')
    db.add_argument('--database', help='MySQL database name')

    pwd = parser.add_argument_group(title='Password')
    pwd.add_argument('-a', '--ask-password', action='store_true',
                     help='Ask password; do not try to retrieve password from keyring')
    pwd.add_argument('-s', '--store-password', action='store_true',
                     help='Store password into keyring after connecting to the database')

    return parser


def get_log_args_parser():
    """
    Parser for the arguments required to set logger level

    :return: an argument parser
    """
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('--log-level', choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
                        help='Verbosity level of the logger')

    return parser


import logging
from getpass import getpass

import keyring
from mysql import connector
from mysql.connector.errors import ProgrammingError

log = logging.getLogger(__name__)


class CursorProvider:
    instance = None

    @classmethod
    def get(cls):
        if CursorProvider.instance is None:
            raise AttributeError('Trying to get instance before instantiating, call {} from {}.{} beforehand'
                                 .format(CursorProvider.init.__name__, __name__, cls.__name__))
        return CursorProvider.instance.connection.cursor()

    @staticmethod
    def init(host, user, port=None, database=None, ask_password=False, store_password=False):
        CursorProvider.instance = CursorProvider.__CursorProvider(host, user, port, database, ask_password,
                                                                  store_password)

    class __CursorProvider:

        def __init__(self, host, user, port=None, database=None, ask_password=False, store_password=False):
            port = port if port is not None else 3306
            log.debug('Trying to connect to the database %s@%s:%s/%s', user, host, port, database)

            service = 'mysql-tracer/{host}'.format(host=host)

            if ask_password:
                self.connect_with_retry(host, port, user, database, service, store_password)
            else:
                log.debug('Retrieving password from keyring (%s@%s)', user, service)
                keyring_password = keyring.get_password(service, user)

                if keyring_password is not None:
                    self.connection = connector.connect(host=host, port=port, user=user, db=database,
                                                        password=keyring_password)
                else:
                    log.info('Did not find password in keyring, asking for password...')
                    self.connect_with_retry(host, port, user, database, service, store_password)

            log.info('Connected to database with success')

        def connect_with_retry(self, host, port, user, database, service, store_password, retry=2):
            password = getpass('Password for {user}@{host}: '.format(user=user, host=host))

            try:
                self.connection = connector.connect(host=host, port=port, user=user, db=database, password=password)
            except ProgrammingError as error:
                if error.errno == 1045 and retry > 0:
                    log.warning('Access Denied, retrying...')
                    self.connect_with_retry(host, port, user, database, service, store_password, retry=retry - 1)
                else:
                    raise error

            if store_password:
                log.info('Storing password into keyring (%s@%s)', user, service)
                keyring.set_password(service, user, password)

        def __del__(self):
            if hasattr(self, 'connection') and self.connection.is_connected():
                log.info('Closing connection to database')
                self.connection.close()



import logging

from mysql_tracer import _configuration
from mysql_tracer._query import Query
from mysql_tracer.args_parser import get_main_args_parser, get_log_args_parser
from mysql_tracer.cursor_provider import CursorProvider

log = logging.getLogger('mysql_tracer')


def configure_logger(log_level):
    """
    Set level of the main logger and add a console handler with the same level

    :param log_level: verbosity level of the logger
    :type log_level: 'DEBUG' | 'INFO' | 'WARNING' | 'ERROR' | 'CRITICAL'
    :return: None
    """
    log.setLevel(log_level)
    console = logging.StreamHandler()
    console.setLevel(log_level)
    console.setFormatter(logging.Formatter(fmt='%(asctime)s %(levelname)s - %(name)s: %(message)s'))
    log.addHandler(console)


def main():
    log_args_parsers = get_log_args_parser()
    log_args, remaining_args = log_args_parsers.parse_known_args()

    if log_args.log_level:
        configure_logger(log_args.log_level)

    config = _configuration.get()

    main_args_parser = get_main_args_parser([log_args_parsers], config)
    main_args = main_args_parser.parse_args()

    if not log_args.log_level and main_args.log_level:
        configure_logger(main_args.log_level)

    CursorProvider.init(main_args.host, main_args.user, main_args.port, main_args.database, main_args.ask_password,
                        main_args.store_password)

    queries = [Query(path) for path in main_args.query]

    for query in queries:
        if main_args.display:
            query.display()
        else:
            query.export(destination=main_args.destination)


if __name__ == '__main__':
    main()



import logging
import re
from datetime import datetime
from os.path import splitext, basename

from mysql_tracer import _writer as writer
from mysql_tracer.cursor_provider import CursorProvider

q_log = logging.getLogger('mysql_tracer.Query')
r_log = logging.getLogger('mysql_tracer.Result')


class Query:
    """
    Represents a MySQL query created from a file. The original file is never modified.

    The reason of existence of this class is to create traces of executed queries so you don't loose track of what you
    did.

    Can execute said query and hold the results through the class Result.
    """

    def __init__(self, source):
        """
        :param source: the path to a file containing a single MySQL statement.
        :type source: str | os.PathLike
        """
        q_log.debug('Initiating Query(source={!r})'.format(source))
        self.source = source
        self.name = splitext(basename(source))[0]
        self.executable_str = self.__executable_str()
        self.result = Result(self.executable_str)
        q_log.debug('Initiated Query(name={!r})'.format(self.name))

    def __repr__(self):
        return 'Query(' \
               'source={source!r}, ' \
               'name={name!r}, ' \
               'executable_str={executable_str!r}, ' \
               'result={result})'.format(**vars(self))

    def __executable_str(self):
        """
        Single line string representation of the query without comments
        """
        query_text = ' '.join([re.sub('(--|#).*', '', line) for line in open(self.source)])
        query_text = re.sub(r'\s+', ' ', query_text)
        return query_text.strip()

    def export(self, destination=None):
        """
        Exports the query to a file with a time prefixed version of the original file name with a
        mini report of the execution appended at the end of the file. And exports the result in a csv file with the same
        name except for the extension.

        :param destination: directory where to create the report and result files
        :type destination: str | os.PathLike
        :return: tuple(report, result)
        :rtype: tuple<str>
        """
        return writer.write(self, destination)

    def display(self):
        print('source: ' + self.source)
        print('sql: ' + self.executable_str)
        print('execution time: {}'.format(self.result.execution_time))
        print('rows count: {}'.format(len(self.result.rows)))
        print('description: {}'.format(self.result.description))
        for row in self.result.rows:
            print(row)


class Result:
    """
    Hold the results of the execution of a MySQL query

    execution_start: datetime before query execution
    execution_end: datetime after query execution
    duration: timedelta of execution_end minus execution_start
    rows: list<tuple<?>> the data the query fetched
    description: tuple<str> the headers of the rows
    """

    def __init__(self, query_str):
        """
        :param query_str: the query to execute
        :type query_str: str
        """
        cursor = CursorProvider.get()
        r_log.info('Executing %s', query_str)
        self.execution_start = datetime.now()
        cursor.execute(query_str)
        self.execution_end = datetime.now()
        self.execution_time = self.execution_end - self.execution_start
        self.rows = cursor.fetchall()
        r_log.debug('Got %d row(s)', len(self.rows))
        self.description = tuple(column[0] for column in cursor.description)

    def __repr__(self):
        return 'Result(' \
               'execution_start={execution_start}, ' \
               'execution_end={execution_end}, ' \
               'execution_time={execution_time}, ' \
               'rows={rows}, ' \
               'description={description})'.format(**vars(self))
